import re
from typing import Dict, Optional, Tuple

from pylabrobot.resources.liquid import Liquid
from pylabrobot.resources.tecan import TipType


def from_str(s: str) -> Optional[Liquid]:
  """Parses a Tecan liquid class name and creates a Liquid object."""

  m = re.match(r"(\w+) free dispense$", s)
  if m is None:
    return None

  lc = m.group(1)
  if lc in {"Ethanol", "Serum"}:
    lc += " 100%"
  return Liquid(lc)


class TecanLiquidClass:
  """A liquid class like used in EVOware."""

  def __init__(
    self,
    lld_mode: int,
    lld_conductivity: int,
    lld_speed: float,
    lld_distance: float,
    clot_speed: float,
    clot_limit: float,
    pmp_sensitivity: int,
    pmp_viscosity: float,
    pmp_character: int,
    density: float,
    calibration_factor: float,
    calibration_offset: float,
    aspirate_speed: float,
    aspirate_delay: float,
    aspirate_stag_volume: float,
    aspirate_stag_speed: float,
    aspirate_lag_volume: float,
    aspirate_lag_speed: float,
    aspirate_tag_volume: float,
    aspirate_tag_speed: float,
    aspirate_excess: float,
    aspirate_conditioning: float,
    aspirate_pinch_valve: bool,
    aspirate_lld: bool,
    aspirate_lld_position: int,
    aspirate_lld_offset: float,
    aspirate_mix: bool,
    aspirate_mix_volume: float,
    aspirate_mix_cycles: int,
    aspirate_retract_position: int,
    aspirate_retract_speed: float,
    aspirate_retract_offset: float,
    dispense_speed: float,
    dispense_breakoff: float,
    dispense_delay: float,
    dispense_tag: bool,
    dispense_pinch_valve: bool,
    dispense_lld: bool,
    dispense_lld_position: int,
    dispense_lld_offset: float,
    dispense_touching_direction: int,
    dispense_touching_speed: float,
    dispense_touching_delay: float,
    dispense_mix: bool,
    dispense_mix_volume: float,
    dispense_mix_cycles: int,
    dispense_retract_position: int,
    dispense_retract_speed: float,
    dispense_retract_offset: float,
  ):
    self.lld_mode = lld_mode
    self.lld_conductivity = lld_conductivity
    self.lld_speed = lld_speed
    self.lld_distance = lld_distance
    self.clot_speed = clot_speed
    self.clot_limit = clot_limit
    self.pmp_sensitivity = pmp_sensitivity
    self.pmp_viscosity = pmp_viscosity
    self.pmp_character = pmp_character
    self.density = density

    self.calibration_factor = calibration_factor
    self.calibration_offset = calibration_offset

    self.aspirate_speed = aspirate_speed
    self.aspirate_delay = aspirate_delay
    self.aspirate_stag_volume = aspirate_stag_volume
    self.aspirate_stag_speed = aspirate_stag_speed
    self.aspirate_lag_volume = aspirate_lag_volume
    self.aspirate_lag_speed = aspirate_lag_speed
    self.aspirate_tag_volume = aspirate_tag_volume
    self.aspirate_tag_speed = aspirate_tag_speed
    self.aspirate_excess = aspirate_excess
    self.aspirate_conditioning = aspirate_conditioning
    self.aspirate_pinch_valve = aspirate_pinch_valve
    self.aspirate_lld = aspirate_lld
    self.aspirate_lld_position = aspirate_lld_position
    self.aspirate_lld_offset = aspirate_lld_offset
    self.aspirate_mix = aspirate_mix
    self.aspirate_mix_volume = aspirate_mix_volume
    self.aspirate_mix_cycles = aspirate_mix_cycles
    self.aspirate_retract_position = aspirate_retract_position
    self.aspirate_retract_speed = aspirate_retract_speed
    self.aspirate_retract_offset = aspirate_retract_offset

    self.dispense_speed = dispense_speed
    self.dispense_breakoff = dispense_breakoff
    self.dispense_delay = dispense_delay
    self.dispense_tag = dispense_tag
    self.dispense_pinch_valve = dispense_pinch_valve
    self.dispense_lld = dispense_lld
    self.dispense_lld_position = dispense_lld_position
    self.dispense_lld_offset = dispense_lld_offset
    self.dispense_touching_direction = dispense_touching_direction
    self.dispense_touching_speed = dispense_touching_speed
    self.dispense_touching_delay = dispense_touching_delay
    self.dispense_mix = dispense_mix
    self.dispense_mix_volume = dispense_mix_volume
    self.dispense_mix_cycles = dispense_mix_cycles
    self.dispense_retract_position = dispense_retract_position
    self.dispense_retract_speed = dispense_retract_speed
    self.dispense_retract_offset = dispense_retract_offset

  def compute_corrected_volume(self, target_volume: float) -> float:
    return self.calibration_factor * target_volume + self.calibration_offset


mapping: Dict[Tuple[float, float, Liquid, TipType], TecanLiquidClass] = {}


def get_liquid_class(
  target_volume: float,
  liquid_class: Liquid,
  tip_type: TipType,
) -> Optional[TecanLiquidClass]:
  for (mnv, mxv, lc, tt), tlc in mapping.items():
    if mnv <= target_volume < mxv and lc == liquid_class and tt == tip_type:
      return tlc
  return None


mapping[(3, 15.01, Liquid.DMSO, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.063,
  calibration_offset=0.1,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=10,
  aspirate_stag_speed=5,
  aspirate_lag_volume=10,
  aspirate_lag_speed=50,
  aspirate_tag_volume=5,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.DMSO, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.1,
  calibration_offset=-0.3,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=5,
  aspirate_lag_volume=0,
  aspirate_lag_speed=50,
  aspirate_tag_volume=10,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.DMSO, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.002,
  calibration_offset=19.3,
  aspirate_speed=150,
  aspirate_delay=300,
  aspirate_stag_volume=20,
  aspirate_stag_speed=5,
  aspirate_lag_volume=0,
  aspirate_lag_speed=50,
  aspirate_tag_volume=10,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.DMSO, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.026,
  calibration_offset=0.1,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=5,
  aspirate_lag_volume=5,
  aspirate_lag_speed=50,
  aspirate_tag_volume=10,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.DMSO, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.007,
  calibration_offset=0.8,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=5,
  aspirate_lag_volume=20,
  aspirate_lag_speed=50,
  aspirate_tag_volume=10,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.DMSO, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.004,
  calibration_offset=3.9,
  aspirate_speed=150,
  aspirate_delay=400,
  aspirate_stag_volume=20,
  aspirate_stag_speed=5,
  aspirate_lag_volume=20,
  aspirate_lag_speed=50,
  aspirate_tag_volume=10,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(0.5, 3.01, Liquid.DMSO, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.311,
  calibration_offset=0,
  aspirate_speed=10,
  aspirate_delay=200,
  aspirate_stag_volume=5,
  aspirate_stag_speed=5,
  aspirate_lag_volume=0,
  aspirate_lag_speed=50,
  aspirate_tag_volume=0.25,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=10,
  dispense_breakoff=10,
  dispense_delay=400,
  dispense_tag=False,
  dispense_pinch_valve=True,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3.01, 15.01, Liquid.DMSO, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.215,
  calibration_offset=1.5,
  aspirate_speed=10,
  aspirate_delay=400,
  aspirate_stag_volume=2,
  aspirate_stag_speed=5,
  aspirate_lag_volume=7,
  aspirate_lag_speed=50,
  aspirate_tag_volume=5,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=110,
  dispense_breakoff=110,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 300.01, Liquid.DMSO, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.122,
  calibration_offset=2.4,
  aspirate_speed=45,
  aspirate_delay=500,
  aspirate_stag_volume=10,
  aspirate_stag_speed=5,
  aspirate_lag_volume=0,
  aspirate_lag_speed=50,
  aspirate_tag_volume=5,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=110,
  dispense_breakoff=110,
  dispense_delay=100,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(1, 7.51, Liquid.DMSO, TipType.MCADITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.35,
  calibration_offset=0,
  aspirate_speed=5,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=5,
  aspirate_lag_volume=7,
  aspirate_lag_speed=50,
  aspirate_tag_volume=2,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=False,
  aspirate_lld_position=0,
  aspirate_lld_offset=0,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=2,
  aspirate_retract_speed=5,
  aspirate_retract_offset=0,
  dispense_speed=500,
  dispense_breakoff=150,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=3,
  dispense_retract_speed=42,
  dispense_retract_offset=0,
)


mapping[(7.51, 20.01, Liquid.DMSO, TipType.MCADITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.04,
  calibration_offset=0,
  aspirate_speed=10,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=5,
  aspirate_lag_volume=7,
  aspirate_lag_speed=50,
  aspirate_tag_volume=3,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=False,
  aspirate_lld_position=0,
  aspirate_lld_offset=0,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=2,
  aspirate_retract_speed=5,
  aspirate_retract_offset=0,
  dispense_speed=500,
  dispense_breakoff=150,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=3,
  dispense_retract_speed=42,
  dispense_retract_offset=0,
)


mapping[(20.01, 200.01, Liquid.DMSO, TipType.MCADITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.04,
  calibration_offset=0,
  aspirate_speed=10,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=5,
  aspirate_lag_volume=10,
  aspirate_lag_speed=50,
  aspirate_tag_volume=3,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=False,
  aspirate_lld_position=0,
  aspirate_lld_offset=0,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=2,
  aspirate_retract_speed=5,
  aspirate_retract_offset=0,
  dispense_speed=400,
  dispense_breakoff=300,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=3,
  dispense_retract_speed=42,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.DMSO, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.08,
  calibration_offset=-0.086,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=50,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=5,
  aspirate_tag_speed=50,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.DMSO, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.024,
  calibration_offset=0.866,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.DMSO, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1.1,
  calibration_factor=1.028,
  calibration_offset=-0.258,
  aspirate_speed=150,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=True,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 1000.01, Liquid.ETHANOL, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.063,
  calibration_offset=3.4,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=0,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 1000.01, Liquid.ETHANOL, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.09,
  calibration_offset=4.3,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=20,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(0.5, 3.01, Liquid.ETHANOL, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.279,
  calibration_offset=0.2,
  aspirate_speed=10,
  aspirate_delay=200,
  aspirate_stag_volume=5,
  aspirate_stag_speed=10,
  aspirate_lag_volume=0,
  aspirate_lag_speed=10,
  aspirate_tag_volume=0.25,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=10,
  aspirate_retract_offset=-5,
  dispense_speed=10,
  dispense_breakoff=10,
  dispense_delay=400,
  dispense_tag=False,
  dispense_pinch_valve=True,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3.01, 15.01, Liquid.ETHANOL, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.199,
  calibration_offset=1.3,
  aspirate_speed=10,
  aspirate_delay=400,
  aspirate_stag_volume=2,
  aspirate_stag_speed=10,
  aspirate_lag_volume=7,
  aspirate_lag_speed=10,
  aspirate_tag_volume=5,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=110,
  dispense_breakoff=110,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 300.01, Liquid.ETHANOL, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.079,
  calibration_offset=2.8,
  aspirate_speed=45,
  aspirate_delay=500,
  aspirate_stag_volume=10,
  aspirate_stag_speed=10,
  aspirate_lag_volume=0,
  aspirate_lag_speed=10,
  aspirate_tag_volume=5,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=110,
  dispense_breakoff=110,
  dispense_delay=100,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 5, Liquid.ETHANOL, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.072,
  calibration_offset=0.22,
  aspirate_speed=20,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=50,
  aspirate_lag_volume=5,
  aspirate_lag_speed=70,
  aspirate_tag_volume=5,
  aspirate_tag_speed=50,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=True,
  aspirate_mix_volume=200,
  aspirate_mix_cycles=2,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(5, 15.01, Liquid.ETHANOL, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.072,
  calibration_offset=0.42,
  aspirate_speed=20,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=50,
  aspirate_lag_volume=5,
  aspirate_lag_speed=70,
  aspirate_tag_volume=5,
  aspirate_tag_speed=50,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=True,
  aspirate_mix_volume=200,
  aspirate_mix_cycles=2,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.ETHANOL, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1.011,
  calibration_offset=1.802,
  aspirate_speed=150,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=5,
  aspirate_lag_speed=70,
  aspirate_tag_volume=5,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=True,
  aspirate_mix_volume=200,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.ETHANOL, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=2,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=0.8,
  calibration_factor=1,
  calibration_offset=0,
  aspirate_speed=150,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=True,
  aspirate_mix_volume=1000,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.SERUM, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.074,
  calibration_offset=0.3,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=10,
  aspirate_stag_speed=20,
  aspirate_lag_volume=20,
  aspirate_lag_speed=20,
  aspirate_tag_volume=10,
  aspirate_tag_speed=20,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 300.01, Liquid.SERUM, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.06,
  calibration_offset=0.43,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=0,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(300.01, 1000.01, Liquid.SERUM, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.007,
  calibration_offset=16.63,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=0,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.SERUM, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.138,
  calibration_offset=1,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=20,
  aspirate_lag_volume=5,
  aspirate_lag_speed=20,
  aspirate_tag_volume=10,
  aspirate_tag_speed=20,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.SERUM, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.059,
  calibration_offset=2.09,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=20,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.SERUM, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.046,
  calibration_offset=8.55,
  aspirate_speed=100,
  aspirate_delay=400,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=20,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.SERUM, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.115,
  calibration_offset=1.146,
  aspirate_speed=50,
  aspirate_delay=600,
  aspirate_stag_volume=0,
  aspirate_stag_speed=50,
  aspirate_lag_volume=15,
  aspirate_lag_speed=70,
  aspirate_tag_volume=5,
  aspirate_tag_speed=50,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-10,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=True,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.SERUM, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.043,
  calibration_offset=2.671,
  aspirate_speed=100,
  aspirate_delay=600,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=20,
  aspirate_lag_speed=70,
  aspirate_tag_volume=6,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-10,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=True,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.SERUM, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=2.1,
  pmp_character=0,
  density=1,
  calibration_factor=1.021,
  calibration_offset=10.966,
  aspirate_speed=100,
  aspirate_delay=600,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=20,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=400,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=True,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.WATER, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.045,
  calibration_offset=0.2,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=10,
  aspirate_stag_speed=20,
  aspirate_lag_volume=10,
  aspirate_lag_speed=20,
  aspirate_tag_volume=5,
  aspirate_tag_speed=20,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 500.01, Liquid.WATER, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.042,
  calibration_offset=-0.04,
  aspirate_speed=150,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=0,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(500.01, 1000.01, Liquid.WATER, TipType.STANDARD)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1,
  calibration_offset=20.26,
  aspirate_speed=150,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=0,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=150,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.WATER, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.042,
  calibration_offset=1.1,
  aspirate_speed=20,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=20,
  aspirate_lag_volume=5,
  aspirate_lag_speed=20,
  aspirate_tag_volume=10,
  aspirate_tag_speed=20,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.WATER, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.034,
  calibration_offset=0.9,
  aspirate_speed=100,
  aspirate_delay=200,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=5,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.WATER, TipType.DITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.031,
  calibration_offset=6.12,
  aspirate_speed=150,
  aspirate_delay=300,
  aspirate_stag_volume=20,
  aspirate_stag_speed=70,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(0.5, 3.01, Liquid.WATER, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.1,
  calibration_offset=0.15,
  aspirate_speed=10,
  aspirate_delay=200,
  aspirate_stag_volume=5,
  aspirate_stag_speed=10,
  aspirate_lag_volume=0,
  aspirate_lag_speed=10,
  aspirate_tag_volume=0.25,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=10,
  dispense_breakoff=10,
  dispense_delay=400,
  dispense_tag=False,
  dispense_pinch_valve=True,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3.01, 15.01, Liquid.WATER, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.045,
  calibration_offset=0.2,
  aspirate_speed=10,
  aspirate_delay=400,
  aspirate_stag_volume=2,
  aspirate_stag_speed=10,
  aspirate_lag_volume=7,
  aspirate_lag_speed=10,
  aspirate_tag_volume=5,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=110,
  dispense_breakoff=110,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 300.01, Liquid.WATER, TipType.STDLOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.06,
  calibration_offset=0.5,
  aspirate_speed=45,
  aspirate_delay=500,
  aspirate_stag_volume=10,
  aspirate_stag_speed=10,
  aspirate_lag_volume=0,
  aspirate_lag_speed=10,
  aspirate_tag_volume=5,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=110,
  dispense_breakoff=110,
  dispense_delay=100,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(1, 3.01, Liquid.WATER, TipType.DITILOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1,
  calibration_offset=0.25,
  aspirate_speed=10,
  aspirate_delay=500,
  aspirate_stag_volume=5,
  aspirate_stag_speed=10,
  aspirate_lag_volume=0,
  aspirate_lag_speed=10,
  aspirate_tag_volume=0.25,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=10,
  dispense_breakoff=10,
  dispense_delay=100,
  dispense_tag=False,
  dispense_pinch_valve=True,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(3.01, 15.01, Liquid.WATER, TipType.DITILOWVOL)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.03,
  calibration_offset=0.25,
  aspirate_speed=10,
  aspirate_delay=500,
  aspirate_stag_volume=10,
  aspirate_stag_speed=10,
  aspirate_lag_volume=10,
  aspirate_lag_speed=10,
  aspirate_tag_volume=3,
  aspirate_tag_speed=10,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=1,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=240,
  dispense_breakoff=110,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(1, 7.51, Liquid.WATER, TipType.MCADITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1,
  calibration_offset=0.45,
  aspirate_speed=5,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=5,
  aspirate_lag_volume=20,
  aspirate_lag_speed=50,
  aspirate_tag_volume=3,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=False,
  aspirate_lld_position=0,
  aspirate_lld_offset=0,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=2,
  aspirate_retract_speed=5,
  aspirate_retract_offset=0,
  dispense_speed=500,
  dispense_breakoff=150,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=3,
  dispense_retract_speed=42,
  dispense_retract_offset=0,
)


mapping[(7.51, 20.01, Liquid.WATER, TipType.MCADITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1,
  calibration_offset=0.45,
  aspirate_speed=10,
  aspirate_delay=200,
  aspirate_stag_volume=0,
  aspirate_stag_speed=5,
  aspirate_lag_volume=10,
  aspirate_lag_speed=50,
  aspirate_tag_volume=5,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=False,
  aspirate_lld_position=0,
  aspirate_lld_offset=0,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=2,
  aspirate_retract_speed=5,
  aspirate_retract_offset=0,
  dispense_speed=500,
  dispense_breakoff=150,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=3,
  dispense_retract_speed=42,
  dispense_retract_offset=0,
)


mapping[(20.01, 200.01, Liquid.WATER, TipType.MCADITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.05,
  calibration_offset=0.4,
  aspirate_speed=50,
  aspirate_delay=500,
  aspirate_stag_volume=0,
  aspirate_stag_speed=5,
  aspirate_lag_volume=20,
  aspirate_lag_speed=50,
  aspirate_tag_volume=10,
  aspirate_tag_speed=5,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=False,
  aspirate_lld_position=0,
  aspirate_lld_offset=0,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=2,
  aspirate_retract_speed=5,
  aspirate_retract_offset=0,
  dispense_speed=400,
  dispense_breakoff=300,
  dispense_delay=200,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=3,
  dispense_retract_speed=42,
  dispense_retract_offset=0,
)


mapping[(3, 15.01, Liquid.WATER, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.05,
  calibration_offset=1.5,
  aspirate_speed=20,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=20,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=5,
  aspirate_tag_speed=20,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(15.01, 200.01, Liquid.WATER, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.03,
  calibration_offset=1.5,
  aspirate_speed=100,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=5,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)


mapping[(200.01, 1000.01, Liquid.WATER, TipType.AIRDITI)] = TecanLiquidClass(
  lld_mode=7,
  lld_conductivity=1,
  lld_speed=60,
  lld_distance=4,
  clot_speed=50,
  clot_limit=4,
  pmp_sensitivity=1,
  pmp_viscosity=1,
  pmp_character=0,
  density=1,
  calibration_factor=1.025,
  calibration_offset=4,
  aspirate_speed=150,
  aspirate_delay=400,
  aspirate_stag_volume=0,
  aspirate_stag_speed=70,
  aspirate_lag_volume=10,
  aspirate_lag_speed=70,
  aspirate_tag_volume=10,
  aspirate_tag_speed=70,
  aspirate_excess=0,
  aspirate_conditioning=0,
  aspirate_pinch_valve=False,
  aspirate_lld=True,
  aspirate_lld_position=3,
  aspirate_lld_offset=2,
  aspirate_mix=False,
  aspirate_mix_volume=100,
  aspirate_mix_cycles=1,
  aspirate_retract_position=4,
  aspirate_retract_speed=20,
  aspirate_retract_offset=-5,
  dispense_speed=600,
  dispense_breakoff=400,
  dispense_delay=0,
  dispense_tag=False,
  dispense_pinch_valve=False,
  dispense_lld=False,
  dispense_lld_position=0,
  dispense_lld_offset=0,
  dispense_touching_direction=0,
  dispense_touching_speed=10,
  dispense_touching_delay=100,
  dispense_mix=False,
  dispense_mix_volume=100,
  dispense_mix_cycles=1,
  dispense_retract_position=1,
  dispense_retract_speed=50,
  dispense_retract_offset=0,
)



import json
import time
import unittest
import unittest.mock
import urllib.request

import pytest
import websockets

from pylabrobot.__version__ import STANDARD_FORM_JSON_VERSION
from pylabrobot.resources import (
  Coordinate,
  Cor_96_wellplate_360ul_Fb,
  Resource,
)
from pylabrobot.visualizer import Visualizer
from pylabrobot.visualizer.visualizer import _sanitize_floats, _serialize_with_methods


class SanitizeFloatsTests(unittest.TestCase):
  def test_inf_replaced(self):
    result = _sanitize_floats({"v": float("inf")})
    self.assertEqual(result, {"v": "Infinity"})
    self.assertEqual(json.dumps(result), '{"v": "Infinity"}')

  def test_neg_inf_replaced(self):
    result = _sanitize_floats({"v": float("-inf")})
    self.assertEqual(result, {"v": "-Infinity"})

  def test_nan_replaced(self):
    result = _sanitize_floats({"v": float("nan")})
    self.assertEqual(result, {"v": "NaN"})

  def test_finite_float_unchanged(self):
    self.assertEqual(_sanitize_floats({"v": 3.14}), {"v": 3.14})

  def test_non_floats_unchanged(self):
    data = {"s": "hello", "i": 42, "b": True, "n": None}
    self.assertEqual(_sanitize_floats(data), data)

  def test_nested_dict(self):
    data = {"a": {"b": {"c": float("inf")}}}
    self.assertEqual(_sanitize_floats(data), {"a": {"b": {"c": "Infinity"}}})

  def test_values_in_list(self):
    data = {"vals": [1.0, float("inf"), float("-inf"), float("nan")]}
    result = _sanitize_floats(data)
    self.assertEqual(result, {"vals": [1.0, "Infinity", "-Infinity", "NaN"]})

  def test_string_containing_infinity_not_touched(self):
    data = {"msg": "status: Infinity reached"}
    self.assertEqual(_sanitize_floats(data), data)

  def test_result_is_valid_json(self):
    data = {"a": float("inf"), "b": [float("-inf")], "c": {"d": float("nan")}}
    serialized = json.dumps(_sanitize_floats(data))
    roundtripped = json.loads(serialized)
    self.assertEqual(roundtripped["a"], "Infinity")
    self.assertEqual(roundtripped["b"], ["-Infinity"])
    self.assertEqual(roundtripped["c"]["d"], "NaN")


class VisualizerLiquidColorValidationTests(unittest.TestCase):
  """Tests for liquid_color parameter validation."""

  def test_valid_hex(self):
    r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    vis = Visualizer(r, open_browser=False, liquid_color="5DADE2")
    self.assertEqual(vis._liquid_color, "5DADE2")

  def test_valid_hex_with_hash(self):
    r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    vis = Visualizer(r, open_browser=False, liquid_color="#aabbcc")
    self.assertEqual(vis._liquid_color, "AABBCC")

  def test_invalid_hex_raises(self):
    r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    with self.assertRaises(ValueError):
      Visualizer(r, open_browser=False, liquid_color="nope")

  def test_short_hex_raises(self):
    r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    with self.assertRaises(ValueError):
      Visualizer(r, open_browser=False, liquid_color="FFF")


class VisualizerSetupStopTests(unittest.IsolatedAsyncioTestCase):
  """Tests for the setup and stop methods of the visualizer backend."""

  @pytest.mark.timeout(20)
  async def test_setup_stop(self):
    """Test that the thread is started and stopped correctly."""

    r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    vis = Visualizer(r, open_browser=False)

    async def setup_stop_single():
      await vis.setup()
      self.assertIsNotNone(vis.loop)
      # wait for the server to start
      time.sleep(1)
      await vis.stop()
      self.assertFalse(vis.has_connection())

    # setup and stop twice to ensure that everything is recycled correctly
    await setup_stop_single()
    await setup_stop_single()


class VisualizerServerTests(unittest.IsolatedAsyncioTestCase):
  """Tests for servers (ws/fs)."""

  async def asyncSetUp(self):
    await super().asyncSetUp()
    self.r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    self.vis = Visualizer(self.r, open_browser=False)
    await self.vis.setup()

    ws_port = self.vis.ws_port  # port may change if port is already in use
    self.uri = f"ws://localhost:{ws_port}"
    self.client = await websockets.connect(self.uri)

  async def asyncTearDown(self):
    await super().asyncTearDown()
    await self.vis.stop()
    await self.client.close()

  def test_get_index_html(self):
    """Test that the index.html file is returned."""
    r = urllib.request.urlopen(f"http://localhost:{self.vis.fs_port}/", timeout=10)
    self.assertEqual(r.status, 200)
    self.assertIn(
      r.headers["Content-Type"],
      ["text/html", "text/html; charset=utf-8"],
    )

  def test_liquid_color_default_substituted(self):
    """Test that the default liquid_color is substituted into the served HTML."""
    url = f"http://localhost:{self.vis.fs_port}/"
    html = urllib.request.urlopen(url, timeout=10).read().decode()
    self.assertNotIn("{{ liquid_color }}", html)
    self.assertIn('value="F39C12"', html)

  async def test_connect(self):
    await self.client.send('{"event": "ready"}')
    response = await self.client.recv()
    response = json.loads(response)
    self.assertEqual(
      response,
      {
        "event": "set_root_resource",
        "data": {
          "resource": _serialize_with_methods(self.r),
        },
        "id": "0001",
        "version": STANDARD_FORM_JSON_VERSION,
      },
    )

  async def test_event_sent(self):
    await self.client.send('{"event": "ready"}')
    _ = await self.client.recv()  # set_root_resource
    _ = await self.client.recv()  # set_state
    _ = await self.client.recv()  # show_machine_tools

    await self.vis.send_command("test", wait_for_response=False)
    recv = await self.client.recv()
    data = json.loads(recv)
    self.assertEqual(data["event"], "test")


class VisualizerShowMachineToolsTests(unittest.IsolatedAsyncioTestCase):
  """Tests for the show_machine_tools_at_start parameter."""

  async def test_show_machine_tools_at_start_false(self):
    """When show_machine_tools_at_start=False, the show_machine_tools event should not be sent."""
    r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    vis = Visualizer(r, open_browser=False, show_machine_tools_at_start=False)
    vis.send_command = unittest.mock.AsyncMock()  # type: ignore[method-assign]
    await vis.setup()

    # Simulate browser ready
    await vis._send_resources_and_state()

    # Check that show_machine_tools was never sent
    for call in vis.send_command.call_args_list:  # type: ignore[attr-defined]
      self.assertNotEqual(
        call[1].get("event") if call[1] else call[0][0],
        "show_machine_tools",
        "show_machine_tools event should not be sent when show_machine_tools_at_start=False",
      )

    await vis.stop()


class VisualizerCommandTests(unittest.IsolatedAsyncioTestCase):
  """Tests for command sending using the visualizer backend."""

  async def asyncSetUp(self):
    await super().asyncSetUp()
    self.maxDiff = None
    self.r = Resource(size_x=100, size_y=100, size_z=100, name="root")
    self.vis = Visualizer(self.r, open_browser=False)

    # mock the send_command method to catch the events
    self.vis.send_command = unittest.mock.AsyncMock()  # type: ignore[method-assign]

    await self.vis.setup()

  async def test_assign_child_resource(self):
    """Test that the assign_child_resource method sends the correct event."""
    child = Resource(size_x=100, size_y=100, size_z=100, name="child")
    self.r.assign_child_resource(child, location=Coordinate(0, 0, 0))
    time.sleep(0.1)  # wait for the event to be sent
    self.vis.send_command.assert_called_once_with(  # type: ignore[attr-defined]
      event="resource_assigned",
      data={
        "resource": _serialize_with_methods(child),
        "state": child.serialize_all_state(),
        "parent_name": "root",
      },
      wait_for_response=False,
    )

  async def test_resource_unassigned(self):
    """Test that the unassign_child_resource method sends the correct event."""
    child = Resource(size_x=100, size_y=100, size_z=100, name="child")
    self.r.assign_child_resource(child, location=Coordinate(0, 0, 0))
    self.r.unassign_child_resource(child)
    time.sleep(0.1)

    self.vis.send_command.assert_called_with(  # type: ignore[attr-defined]
      event="resource_unassigned",
      data={"resource_name": "child"},
      wait_for_response=False,
    )

  async def test_state_updated(self):
    """Test that the state_updated method sends the correct event."""
    plate = Cor_96_wellplate_360ul_Fb(name="plate_01")
    self.r.assign_child_resource(plate, location=Coordinate(0, 0, 0))
    plate.set_well_volumes([500] * 96)
    time.sleep(0.1)
    self.vis.send_command.assert_called()  # type: ignore[attr-defined]
    call_args = self.vis.send_command.call_args[1]  # type: ignore[attr-defined]
    self.assertEqual(call_args["event"], "set_state")
    self.assertEqual(
      call_args["data"]["plate_01_well_H12"]["volume"],
      500,
    )



# originally written by Delf Byonoy
# https://github.com/byonoy-dne/report-parsing/blob/main/parser.py

import re
import struct

# Map report_id to the corresponding payload structure
PAYLOAD_FORMATS = {
  0x0010: (
    "<BB29H",
    "SUPPORTED_REPORTS_IN",
    ["seq", "seq_len", "ids"],
  ),  # supported_reports_in_t
  0x0050: ("<I", "API_VERSION_IN", ["version_no"]),  # api_version_in_t
  0x0020: (
    "<HH56s",
    "ACK_INOUT",
    ["report_id", "return_code", "error_description"],
  ),  # ack_inout_t
  0x0500: (
    "<BBhhI12fBB",
    "ABS96_MEASUREMENT_IN",
    [
      "seq",
      "seq_len",
      "signal_wavelength_nm",
      "reference_wavelength_nm",
      "duration_ms",
      "data",
      "flags",
      "progress",
    ],
  ),  # abs96_measurement_in_t
  0x0300: (
    "<BBBIBB",
    "STATUS_IN",
    [
      "is_initialized",
      "is_slot_occupied",
      "error_code",
      "time_since_boot_s",
      "is_measurement_in_progress",
      "boot_completed",
    ],
  ),  # status_in_t
  0x0070: (
    "<BII50s",
    "DIAGNOSTIC_DATA_IN",
    ["level", "offset", "data_length", "block"],
  ),  # diagnostic_data_in_t
  0x0030: (
    "<II52s",
    "BULK_DATA_INOUT",
    ["offset", "data_length", "block"],
  ),  # bulk_data_inout_t
  0x0200: (
    "<HB52s",
    "DEVICE_DATA_READ_IN",
    ["field_index", "flags", "data"],
  ),  # device_data_read_in_t
  0x0210: (
    "<HB52s",
    "DEVICE_DATA_WRITE_OUT",
    ["field_index", "flags", "data"],
  ),  # device_data_write_out_t
  0x0220: (
    "<HH55sB",
    "DEVICE_DATA_FIELDS_IN",
    ["field_index", "field_count", "field_name", "flags"],
  ),  # device_data_fields_in_t
  0x0240: (
    "<HBI",
    "DEVICE_FILE_READ_IN",
    ["file_index", "flags", "file_size"],
  ),  # device_file_read_in_t
  0x0250: (
    "<HBI",
    "DEVICE_FILE_WRITE_OUT",
    ["file_index", "flags", "file_size"],
  ),  # device_file_write_out_t
  0x0260: (
    "<HH51sIBx",  # Added 1 byte padding to align with struct
    "DEVICE_FILE_NAMES_IN",
    ["file_index", "file_count", "file_name", "file_size", "flags"],
  ),  # device_file_names_in_t
  0x0280: (
    "<HH55sB",
    "RPC_ENDPOINTS_IN",
    ["rpc_index", "rpc_count", "function_name", "flags"],
  ),  # rpc_endpoints_in_t
  0x0290: (
    "<H",
    "RPC_EXECUTE_OUT",
    ["rpc_index"],
  ),  # rpc_execute_out_t
  0x0320: (
    "<hhBB",
    "ABS_TRIGGER_MEASUREMENT_OUT",
    [
      "signal_wavelength_nm",
      "reference_wavelength_nm",
      "is_reference_measurement",
      "flags",
    ],
  ),  # abs_trigger_measurement_out_t
  0x0330: (
    "<30h",
    "ABS_WAVELENGTHS_IN",
    ["wavelengths_nm"],
  ),  # abs_wavelengths_in_t
  0x0340: (
    "<i12sBB",
    "LUM_TRIGGER_MEASUREMENT_OUT",
    [
      "integration_time_us",
      "channels_selected",
      "is_reference_measurement",
      "flags",
    ],
  ),  # lum_trigger_measurement_out_t
  0x0341: (
    "<i48sBB",
    "LUM384_TRIGGER_MEASUREMENT_OUT",
    [
      "integration_time_us",
      "channels_selected",
      "is_reference_measurement",
      "flags",
    ],
  ),  # lum384_trigger_measurement_out_t
  0x0600: (
    "<BBII12fBB",
    "LUM96_MEASUREMENT_IN",
    [
      "seq",
      "seq_len",
      "integration_time_us",
      "duration_ms",
      "row",
      "flags",
      "progress",
    ],
  ),  # lum96_measurement_in_t
  0x0610: (
    "<BBII12fBB",
    "LUM384_MEASUREMENT_IN",
    [
      "seq",
      "seq_len",
      "integration_time_us",
      "duration_ms",
      "row",
      "flags",
      "progress",
    ],
  ),  # lum384_measurement_in_t
  0x1000: (
    "<BBBIBIBh",
    "ESP_STATUS_IN",
    [
      "connection_state",
      "last_connection_error",
      "is_connected_to_cloud",
      "last_cloud_connection_error_code",
      "self_test_status",
      "self_test_errors",
      "wifi_phy_mode",
      "wifi_rssi",
    ],
  ),  # esp_status_in_t
  0x0040: (
    "<B",
    "HEARTBEAT_IN",
    ["enabled"],
  ),  # heartbeat_in_t
}

# Create a reverse mapping from payload name to report ID for easier encoding
REPORT_IDS = {
  payload_name: report_id for report_id, (_, payload_name, _) in PAYLOAD_FORMATS.items()
}


def parse_hid_report(binary_data):
  # Parse the outer hid_report_t structure
  header_fmt = "<H"  # Little-endian: H=uint16 for report_id
  header_size = struct.calcsize(header_fmt)
  (report_id,) = struct.unpack(header_fmt, binary_data[:header_size])

  if report_id not in PAYLOAD_FORMATS:
    raise ValueError(f"Unknown report_id: 0x{report_id:04x}")

  payload_fmt, payload_name, field_names = PAYLOAD_FORMATS[report_id]
  payload_size = struct.calcsize(payload_fmt)
  payload_data = struct.unpack(payload_fmt, binary_data[header_size : header_size + payload_size])

  result = {
    "report_id": report_id,
    "payload_name": payload_name,
    "payload": {field: value for field, value in zip(field_names, payload_data)},
  }

  return result


def encode_hid_report(report_dict):
  report_id = report_dict.get("report_id")
  if report_id not in PAYLOAD_FORMATS:
    raise ValueError(f"Unknown report_id: 0x{report_id:04x}")

  payload_fmt, payload_name, field_names = PAYLOAD_FORMATS[report_id]

  # Prepare the payload data in the correct order, zeroing any fields that
  # were not explicitly provided.
  payload = report_dict.get("payload", {})

  # Find all format characters (e.g., 'i', 's', 'B') in the format string.
  format_chars = re.findall(r"\d*([a-zA-Z])", payload_fmt)

  payload_data = []
  for i, field in enumerate(field_names):
    default = b"" if format_chars[i] == "s" else 0
    p = payload.get(field, default)
    if isinstance(p, list):
      payload_data.extend(p)
    else:
      payload_data.append(p)

  # Encode the payload
  binary_payload = struct.pack(payload_fmt, *payload_data)

  # Encode the full report (header + payload)
  header_fmt = "<H"
  binary_header = struct.pack(header_fmt, report_id)
  packet = binary_header + binary_payload
  packet += b"\x00" * (64 - len(packet))

  return packet



"""Data structures for the standard form of liquid handling."""

from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional, Sequence, Union

from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.rotation import Rotation

if TYPE_CHECKING:
  from pylabrobot.resources import (
    Container,
    Resource,
    TipRack,
    Trash,
    Well,
  )
  from pylabrobot.resources.tip import Tip
  from pylabrobot.resources.tip_rack import TipSpot


@dataclass(frozen=True)
class Pickup:
  resource: TipSpot
  offset: Coordinate
  tip: Tip  # TODO: perhaps we can remove this, because the tip spot has the tip?


@dataclass(frozen=True)
class Drop:
  resource: Resource
  offset: Coordinate
  tip: Tip


@dataclass(frozen=True)
class PickupTipRack:
  resource: TipRack
  offset: Coordinate
  tips: Sequence[Optional[Tip]]


@dataclass(frozen=True)
class DropTipRack:
  resource: Union[TipRack, Trash]
  offset: Coordinate


@dataclass(frozen=True)
class SingleChannelAspiration:
  resource: Container
  offset: Coordinate
  tip: Tip
  volume: float
  flow_rate: Optional[float]
  liquid_height: Optional[float]
  blow_out_air_volume: Optional[float]
  mix: Optional[Mix]


@dataclass(frozen=True)
class SingleChannelDispense:
  resource: Container
  offset: Coordinate
  tip: Tip
  volume: float
  flow_rate: Optional[float]
  liquid_height: Optional[float]
  blow_out_air_volume: Optional[float]
  mix: Optional[Mix]


@dataclass(frozen=True)
class Mix:
  volume: float
  repetitions: int
  flow_rate: float


@dataclass(frozen=True)
class MultiHeadAspirationPlate:
  wells: List[Well]
  offset: Coordinate
  tips: Sequence[Optional[Tip]]
  volume: float
  flow_rate: Optional[float]
  liquid_height: Optional[float]
  blow_out_air_volume: Optional[float]
  mix: Optional[Mix]


@dataclass(frozen=True)
class MultiHeadDispensePlate:
  wells: List[Well]
  offset: Coordinate
  tips: Sequence[Optional[Tip]]
  volume: float
  flow_rate: Optional[float]
  liquid_height: Optional[float]
  blow_out_air_volume: Optional[float]
  mix: Optional[Mix]


@dataclass(frozen=True)
class MultiHeadAspirationContainer:
  container: Container
  offset: Coordinate
  tips: Sequence[Optional[Tip]]
  volume: float
  flow_rate: Optional[float]
  liquid_height: Optional[float]
  blow_out_air_volume: Optional[float]
  mix: Optional[Mix]


@dataclass(frozen=True)
class MultiHeadDispenseContainer:
  container: Container
  offset: Coordinate
  tips: Sequence[Optional[Tip]]
  volume: float
  flow_rate: Optional[float]
  liquid_height: Optional[float]
  blow_out_air_volume: Optional[float]
  mix: Optional[Mix]


class GripDirection(enum.Enum):
  FRONT = enum.auto()
  BACK = enum.auto()
  LEFT = enum.auto()
  RIGHT = enum.auto()


@dataclass(frozen=True)
class ResourcePickup:
  resource: Resource
  offset: Coordinate
  pickup_distance_from_top: float
  direction: GripDirection


@dataclass(frozen=True)
class ResourceMove:
  """Moving a resource that was already picked up."""

  resource: Resource
  location: Coordinate
  gripped_direction: GripDirection
  pickup_distance_from_top: float
  offset: Coordinate


@dataclass(frozen=True)
class ResourceDrop:
  resource: Resource
  # Destination is the location of the lfb of `resource`
  destination: Coordinate
  destination_absolute_rotation: Rotation
  offset: Coordinate
  pickup_distance_from_top: float
  pickup_direction: GripDirection
  direction: GripDirection
  rotation: float


PipettingOp = Union[Pickup, Drop, SingleChannelAspiration, SingleChannelDispense]



from typing import Optional

from pylabrobot.machines.machine import Machine
from pylabrobot.resources import Resource, Rotation
from pylabrobot.scales.scale_backend import ScaleBackend


class Scale(Resource, Machine):
  """Base class for a scale"""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    backend: ScaleBackend,
    rotation: Optional[Rotation] = None,
    category: Optional[str] = None,
    model: Optional[str] = None,
  ):
    Machine.__init__(self, backend=backend)
    Resource.__init__(
      self,
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      rotation=rotation,
      category=category,
      model=model,
    )
    self.backend: ScaleBackend = backend  # fix type

  async def tare(self, **backend_kwargs):
    """Tare the scale"""
    await self.backend.tare(**backend_kwargs)

  async def zero(self, **backend_kwargs):
    """Zero the scale"""
    await self.backend.zero(**backend_kwargs)

  async def get_weight(self, **backend_kwargs) -> float:
    """Get the weight in grams"""
    return await self.backend.get_weight(**backend_kwargs)



from typing import cast

from pylabrobot.temperature_controlling.backend import (
  TemperatureControllerBackend,
)

try:
  import ot_api

  USE_OT = True
except ImportError as e:
  USE_OT = False
  _OT_IMPORT_ERROR = e


class OpentronsTemperatureModuleBackend(TemperatureControllerBackend):
  """Opentrons temperature module backend."""

  @property
  def supports_active_cooling(self) -> bool:
    return False

  def __init__(self, opentrons_id: str):
    """Create a new Opentrons temperature module backend.

    Args:
      opentrons_id: Opentrons ID of the temperature module. Get it from
        `OpentronsBackend(host="x.x.x.x", port=31950).list_connected_modules()`.
    """
    self.opentrons_id = opentrons_id

    if not USE_OT:
      raise RuntimeError(
        "Opentrons is not installed. Please run pip install pylabrobot[opentrons]."
        f" Import error: {_OT_IMPORT_ERROR}."
      )

  async def setup(self):
    pass

  async def stop(self):
    await self.deactivate()

  def serialize(self) -> dict:
    return {**super().serialize(), "opentrons_id": self.opentrons_id}

  async def set_temperature(self, temperature: float):
    ot_api.modules.temperature_module_set_temperature(
      celsius=temperature, module_id=self.opentrons_id
    )

  async def deactivate(self):
    ot_api.modules.temperature_module_deactivate(module_id=self.opentrons_id)

  async def get_current_temperature(self) -> float:
    modules = ot_api.modules.list_connected_modules()
    for module in modules:
      if module["id"] == self.opentrons_id:
        return cast(float, module["data"]["currentTemperature"])
    raise RuntimeError(f"Module with id '{self.opentrons_id}' not found")



import enum
import math
from typing import Callable, Dict, List, Optional, Tuple, Union

from pylabrobot.resources.container import Container
from pylabrobot.resources.liquid import Liquid
from pylabrobot.resources.plate import Plate


class WellBottomType(enum.Enum):
  """Enum for the type of bottom of a well."""

  FLAT = "flat"
  U = "U"
  V = "V"
  UNKNOWN = "unknown"


class CrossSectionType(enum.Enum):
  """Enum for the type of cross section of a well.

  A well with a circular cross section will be a cylinder, and a well with a rectangular cross
  section will be a cuboid. Note that the bottom section of a well may be any of the
  :class:`WellBottomType` values.
  """

  CIRCLE = "circle"
  RECTANGLE = "rectangle"


class Well(Container):
  """Base class for Well resources.

  Note that in regular use these will be automatically generated by the
  :class:`pylabrobot.resources.Plate` class.
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    material_z_thickness: Optional[float] = None,
    bottom_type: Union[WellBottomType, str] = WellBottomType.UNKNOWN,
    category: str = "well",
    model: Optional[str] = None,
    max_volume: Optional[float] = None,
    compute_volume_from_height: Optional[Callable[[float], float]] = None,
    compute_height_from_volume: Optional[Callable[[float], float]] = None,
    cross_section_type: Union[CrossSectionType, str] = CrossSectionType.CIRCLE,
    height_volume_data: Optional[Dict[float, float]] = None,
  ):
    """Create a new well.

    Args:
      name: Name of the well.
      size_x: Size of the well in the x direction.
      size_y: Size of the well in the y direction.
      size_z: Size of the well in the z direction.
      bottom_type: Type of the bottom of the well. If a string, must be the raw value of the
        :class:`WellBottomType` enum. This is used to deserialize and may be removed in the future.
      category: Category of the well.
      max_volume: Maximum volume of the well. If not specified, the well will be seen as a cylinder
        and the max volume will be computed based on size_x, size_y, and size_z.
      compute_volume_from_height: function to compute the volume from the height relative to the
        bottom
      cross_section_type: Type of the cross section of the well. If not specified, the well will be
        seen as a cylinder.
      height_volume_data: Optional dict mapping height (mm) to volume (uL). See
        :class:`Container` for details.
    """

    if isinstance(bottom_type, str):
      bottom_type = WellBottomType(bottom_type)
    if isinstance(cross_section_type, str):
      cross_section_type = CrossSectionType(cross_section_type)

    if max_volume is None:
      if compute_volume_from_height is None:
        # we assume flat bottom as a best guess, bottom types require additional information
        if cross_section_type == CrossSectionType.CIRCLE:
          assert size_x == size_y, "size_x and size_y must be equal for circular wells."
          max_volume = math.pi * (size_x / 2) ** 2 * size_z
        elif cross_section_type == CrossSectionType.RECTANGLE:
          max_volume = size_x * size_y * size_z
      else:
        max_volume = compute_volume_from_height(size_z)

    super().__init__(
      name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      category=category,
      max_volume=max_volume,
      model=model,
      compute_volume_from_height=compute_volume_from_height,
      compute_height_from_volume=compute_height_from_volume,
      material_z_thickness=material_z_thickness,
      height_volume_data=height_volume_data,
    )
    self.bottom_type = bottom_type
    self.cross_section_type = cross_section_type

    self.tracker.register_callback(self._state_updated)

  def serialize(self):
    return {
      **super().serialize(),
      "bottom_type": self.bottom_type.value,
      "cross_section_type": self.cross_section_type.value,
    }

  def set_volume(self, volume: float):
    """Set the volume of the well.

    (wraps :meth:`~.VolumeTracker.set_volume`)

    Example:
      Set the volume in a well to 10 uL:

      >>> well.set_volume(10)
    """

    self.tracker.set_volume(volume)

  def set_liquids(self, liquids: List[Tuple[Optional["Liquid"], float]]):
    """Set the liquids in the well.

    Deprecated: Use `set_volume` instead. This method will be removed in a future version.
    """

    self.tracker.set_liquids(liquids)

  def _get_parent_plate(self) -> Plate:
    if self.parent is None or not isinstance(self.parent, Plate):
      raise ValueError("Well must be in a plate to get its parent plate.")
    return self.parent

  def get_identifier(self) -> str:
    """Get the (canonical) identifier, like `"A1"` of the well in the parent plate. If the well is not in a plate, this will raise a ValueError."""
    return self._get_parent_plate().get_child_identifier(self)

  def get_row(self) -> int:
    """Get the row (0-indexed) of the well in the parent plate. If the well is not in a plate, this will raise a ValueError."""
    return self._get_parent_plate().get_child_row(self)

  def get_column(self) -> int:
    """Get the column (0-indexed) of the well in the parent plate. If the well is not in a plate, this will raise a ValueError."""
    return self._get_parent_plate().get_child_column(self)



"""EL406 protocol framing utilities.

This module contains the protocol framing functions for building
properly formatted messages for the BioTek EL406 plate washer.
"""

from __future__ import annotations

from pylabrobot.io.binary import Writer


def build_framed_message(command: int, data: bytes = b"") -> bytes:
  """Build a properly framed EL406 message.

  Protocol structure:
    [0]: 0x01 (start marker)
    [1]: 0x02 (version marker)
    [2-3]: command (little-endian short)
    [4]: 0x01 (constant)
    [5-6]: reserved (ushort, typically 0)
    [7-8]: data length (ushort, little-endian)
    [9-10]: checksum (ushort, little-endian)
    ... followed by data bytes

  Checksum is two's complement of sum of header bytes 0-8 + all data bytes.

  Args:
    command: 16-bit command code
    data: Optional data bytes

  Returns:
    Complete framed message with header and checksum
  """
  # Build header bytes 0-8 (checksum placeholder filled after)
  header_prefix = (
    Writer()
    .u8(0x01)      # [0] Start marker
    .u8(0x02)    # [1] Version marker
    .u16(command)              # [2-3] Command (LE)
    .u8(0x01)          # [4] Constant
    .u16(0x0000)               # [5-6] Reserved
    .u16(len(data))            # [7-8] Data length (LE)
    .finish()
  )  # fmt: skip

  # Checksum: two's complement of sum of header bytes 0-8 + all data bytes
  checksum_sum = sum(header_prefix) + sum(data)
  checksum = (0xFFFF - checksum_sum + 1) & 0xFFFF

  return header_prefix + Writer().u16(checksum).finish() + data


def encode_column_mask(columns: list[int] | None) -> bytes:
  """Encode list of column indices to 6-byte (48-bit) column mask.

  Each bit represents one column: 0 = skip, 1 = operate on column.

  Args:
    columns: List of column indices (0-47) to select, or None for all columns.
      If None, returns all 1s (all columns selected).
      If empty list, returns all 0s (no columns selected).

  Returns:
    6 bytes representing the 48-bit column mask in little-endian order.

  Raises:
    ValueError: If any column index is out of range (not 0-47).
  """
  if columns is None:
    return bytes([0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF])

  for col in columns:
    if col < 0 or col > 47:
      raise ValueError(f"Column index {col} out of range. Must be 0-47.")

  mask = [0] * 6
  for col in columns:
    byte_index = col // 8
    bit_index = col % 8
    mask[byte_index] |= 1 << bit_index

  return bytes(mask)


def columns_to_column_mask(columns: list[int] | None, plate_wells: int = 96) -> list[int] | None:
  """Convert 1-indexed column numbers to 0-indexed column indices.

  Args:
    columns: List of column numbers (1-based), or None for all columns.
    plate_wells: Plate format (96, 384, 1536). Determines max columns.

  Returns:
    List of 0-indexed column indices, or None if columns is None.

  Raises:
    ValueError: If column numbers are out of range.
  """
  if columns is None:
    return None

  max_cols = {96: 12, 384: 24, 1536: 48}.get(plate_wells, 48)
  indices = []
  for col in columns:
    if col < 1 or col > max_cols:
      raise ValueError(f"Column {col} out of range for {plate_wells}-well plate (1-{max_cols}).")
    indices.append(col - 1)
  return indices



import warnings

from pylabrobot.resources.carrier import (
  Coordinate,
  ResourceHolder,
  TubeCarrier,
  create_homogeneous_resources,
)


def Tube_CAR_24_A00(name: str) -> TubeCarrier:
  """Hamilton cat. no.: 173400
  Hamilton name: 'SMP_CAR_24_A00'.
  'Sample' carrier for 24 tubes sizes 14.5x60 - 18x120mm.
  1 track(T) wide.
  """
  return TubeCarrier(
    name=name,
    size_x=22.5,
    size_y=497.0,
    size_z=71.5,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(3.0, 9.0 + x * 20, 10.0 + 1.2) for x in range(24)
      ],  # TODO: +1.2 to account for the Tube.material_z_thickness, fix container
      resource_size_x=18.0,
      resource_size_y=18.0,
      name_prefix=name,
    ),
    model="Tube_CAR_24_A00",
  )


def Tube_CAR_32_A00(name: str) -> TubeCarrier:
  warnings.warn(
    "The true dimensions of a Tube_CAR_32_A00 (SMP_CAR_32_A00) are not known. "
    "The hamilton definitions are with inserts. "
    "Do you want to use `hamilton_tube_carrier_32_a00_insert_eppendorf_1_5mL` instead?",
    DeprecationWarning,
  )
  return hamilton_tube_carrier_32_a00_insert_eppendorf_1_5mL(name)


def hamilton_tube_carrier_32_a00_insert_eppendorf_1_5mL(name: str) -> TubeCarrier:
  """Hamilton cat. no.: 173410 with inserts cat no. 187350.
  Hamilton name: 'SMP_CAR_32_A00'.
  'Sample' carrier for 32 tubes
  1 track(T) wide.
  For use with `Eppendorf_DNA_LoBind_1_5ml_Vb` and `Eppendorf_Protein_LoBind_1_5ml_Vb`.
  """
  hole_diameter = 10.8
  return TubeCarrier(
    name=name,
    size_x=22.5,  # 1 track
    size_y=497.0,  # standard
    size_z=60.2,  # caliper
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(14.5 - hole_diameter / 2, 14.5 - hole_diameter / 2 + x * 15, 25.88)
        for x in range(
          32
        )  # SMP_CAR_32_A00 in venus, verified with caliper, custom z from z probing
      ],
      # should fix container
      resource_size_x=hole_diameter,  # venus
      resource_size_y=hole_diameter,  # venus
      name_prefix=name,
    ),
    model=hamilton_tube_carrier_32_a00_insert_eppendorf_1_5mL.__name__,
  )


def hamilton_tube_carrier_12_b00(name: str) -> TubeCarrier:
  """Hamilton cat. no.: 182045
  Hamilton name: 'SMP_CAR_12_B00'.
  'Sample' carrier for 12 50mL falcon tubes (Cor_Falcon_tube_50mL_Vb).
  2 track(T) wide.
  """
  hole_diameter = 29.0
  return TubeCarrier(
    name=name,
    size_x=45,  # 2 tracks
    size_y=497.0,  # standard
    size_z=92.0,  # caliper
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(26.05 - hole_diameter / 2, 41.74 - hole_diameter / 2 + i * 36.5, 18.9)
        for i in range(12)  # SMP_CAR_12_A00 in venus, verified with caliper, custom z
      ],
      resource_size_x=hole_diameter,
      resource_size_y=hole_diameter,
      name_prefix=name,
    ),
    model=hamilton_tube_carrier_12_b00.__name__,
  )



from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.temperature_controlling.inheco.control_box import InhecoTECControlBox
from pylabrobot.temperature_controlling.inheco.cpac_backend import InhecoCPACBackend
from pylabrobot.temperature_controlling.temperature_controller import TemperatureController


def inheco_cpac_ultraflat(
  name: str, control_box: InhecoTECControlBox, index: int
) -> TemperatureController:
  """Inheco CPAC Ultraflat
  7000166, 7000190, 7000165

  https://www.inheco.com/data/pdf/cpac-brochure-1013-1032-34.pdf
  """

  return TemperatureController(
    name=name,
    backend=InhecoCPACBackend(control_box=control_box, index=index),
    size_x=113,  # from spec
    size_y=89,  # from spec
    size_z=129,  # from spec
    child_location=Coordinate(x=8, y=11, z=77),  # x from spec, y and z measured
    model=inheco_cpac_ultraflat.__name__,
  )
